import raf from 'raf'
import React from 'react'

class Store {

  constructor(name, actions, state) {

    this.name = name
    this.actions = actions
    this.initialState = state
    this.state = state
    this.dirtyState = undefined
    this.willChange = undefined

    let handlers = []

    this.commit = () => {
      this.willChange = false
      this.state = this.dirtyState
      this.dirtyState = undefined
      this.didChange()
    }

    this.request = this.request.bind(this)

    this.didChange = () => {

      this.mountStateIfNeeded(this.state)

      handlers.forEach((handler) => {
        handler(this)
      })
    }

    this.mountedComponent = undefined

    this.lastMountedState = undefined

    this.mountStateIfNeeded = (state) => {

      let mounted = this.mountedComponent

      // check if equals b/c mounted component setsState immediately on willReceiveProps
      if (mounted) {

        if (this.lastMountedState && state.equals( this.lastMountedState )) return

        this.lastMountedState = state

        mounted.setState({
          [name]: state
        })

      }

    }

    this.defaultActions = {

      'connect': (state, handler, isDeferred = true) => {
        if (typeof handler !== 'function') throw new Error(`Contextual.Store ${name} connect action accepts a funcion not ${handler}`)

        if (!handlers.includes(handler)) {

          handlers.push( handler )

          // call handler if just adding
          if (!isDeferred) {
            handler(this)
          }
          raf( () => {
            handler(this)
          })

        }

        return state
      },

      'disconnect': (state, handler) => {
        let index = handlers.indexOf(handler)

        if (index === -1) throw new Error(`Contextual.Store ${name} disconnect action cant find ${handler}, you're probably returning new instance of function :)`)

        handlers.splice(index, 1)

        return state
      },

      'get-state': (state, callback) => {
        callback(state)
        return state
      },

      'replace-state': (state, newState) => {
        if (state.equals(newState)) return state
        return newState
      },

      'merge': (state, ...args) => {
        if (args.length === 1) {
          let mergeable = args[0]
          // check if needed...
          if (mergeable.equals && mergeable.equals(state)) return state
          return state.merge(mergeable)
        }
        return state.merge(...args)
      },

      'merge-in': (state, ...args) => {
        return state.mergeIn(...args)
      },

      'merge-deep': (state, ...args) => {
        if (args.length === 1) {
          let mergeable = args[0]
          // check if needed...
          if (mergeable.equals && mergeable.toJS) {
            if (mergeable.equals(state)) return state
            return state.merge(mergeable)
          }
        }
        throw new Error(`Contextual.Store ${name} merge-deep action only accepts one arg of type Immutable`)
      },

      'mount': (state, comp) => {
        if (this.mountedComponent) throw new Error("Store cant have more than one mounted component")

        this.mountedComponent = comp

        let {
          componentWillReceiveProps,
          componentWillUnmount,
          props,
        } = comp

        comp.componentWillReceiveProps = (...args) => {
          if (componentWillReceiveProps) componentWillReceiveProps.apply(comp, args)
          let [nextProps] = args

          let newState = this.state.merge(nextProps)
          if (newState.equals( this.state )) return
          this._setState(newState)

          // setState immediately...
          this.mountStateIfNeeded(newState)
        }

        comp.componentWillUnmount = (...args) => {
          if (componentWillUnmount) componentWillUnmount.apply(comp, args)
          this.lastMountedState = undefined
          this.mountedComponent = undefined
        }

        if (props) state = state.merge(props)

        // setState immediately...
        this.mountStateIfNeeded(state)

        return state
        //return state.merge()
      },

    }
  }

  mount(componentInstance) {
    return this.request('mount', componentInstance)
  }

  getCurrentState() {
    return this.dirtyState ? this.dirtyState : this.state
  }

  request(actionName, ...payload) {

    let action = this.actions[actionName]
    if (!action) action = this.defaultActions[actionName]

    if (!action) throw new Error(`Contextual.Store ${name} does not provide action: ${actionName}` )

    let startState = this.getCurrentState()

    // action methods called with this === store
    return this._setState( action.call( this, startState, ...payload ) )

  }

  _setState(state) {
    this.dirtyState = state

    if (!this.state.equals( this.dirtyState )){

      if (this.willChange) return

      this.willChange = true

      // defering commit
      raf(this.commit)

    }
    else {
      if (this.willChange) raf.cancel(this.commit)
      this.dirtyState = undefined
      this.willChange = false
    }
  }
}

export default Store
